﻿using EonaCat.VolumeMixer.Helpers;
using EonaCat.VolumeMixer.Interfaces;
using System;
using System.Collections.Generic;
using System.Threading.Tasks;

namespace EonaCat.VolumeMixer.Models
{
    // This file is part of the EonaCat project(s) which is released under the Apache License.
    // See the LICENSE file or go to https://EonaCat.com/License for full license details.
    public class AudioDevice : IDisposable
    {
        private readonly IMultiMediaDevice _device;
        private IAudioEndpointVolume _endpointVolume;
        private IAudioSessionManager _sessionManager;
        private bool _isDisposed = false;
        private readonly object _syncLock = new object();

        public string Id { get; private set; }
        public string Name { get; private set; }
        public bool IsDefault { get; private set; }
        public DataFlow DataFlow { get; private set; }

        internal AudioDevice(IMultiMediaDevice device, string id, string name, bool isDefault, DataFlow dataFlow)
        {
            _device = device;
            Id = id;
            Name = name;
            IsDefault = isDefault;
            DataFlow = dataFlow;
            InitializeEndpointVolume();
            InitializeSessionManager();
        }

        private void InitializeEndpointVolume()
        {
            try
            {
                var audioEndpointGuid = typeof(IAudioEndpointVolume).GUID;
                var result = _device.Activate(ref audioEndpointGuid, 0, IntPtr.Zero, out var ptr);
                if (result == 0 && ptr != IntPtr.Zero)
                {
                    _endpointVolume = ComHelper.GetInterface<IAudioEndpointVolume>(ptr);
                }
            }
            catch
            {
                _endpointVolume = null;
            }
        }

        private void InitializeSessionManager()
        {
            try
            {
                var audioSessionGuid = typeof(IAudioSessionManager).GUID;
                var result = _device.Activate(ref audioSessionGuid, 0, IntPtr.Zero, out var ptr);
                if (result == 0 && ptr != IntPtr.Zero)
                {
                    _sessionManager = ComHelper.GetInterface<IAudioSessionManager>(ptr);
                }
            }
            catch
            {
                _sessionManager = null;
            }
        }

        public async Task<float> GetMasterVolumeAsync()
        {
            return await Task.Run(() =>
            {
                lock (_syncLock)
                {
                    if (_isDisposed || _endpointVolume == null)
                    {
                        return 0f;
                    }

                    try
                    {
                        var result = _endpointVolume.GetMasterVolumeLevelScalar(out var volume);
                        return result == 0 ? volume : 0f;
                    }
                    catch
                    {
                        return 0f;
                    }
                }
            });
        }

        public async Task<bool> SetMasterVolumeAsync(float volume, int maxRetries = 5, int delayMs = 20)
        {
            for (int attempt = 0; attempt <= maxRetries; attempt++)
            {
                lock (_syncLock)
                {
                    if (_isDisposed || _endpointVolume == null || volume < 0f || volume > 1f)
                    {
                        return false;
                    }

                    try
                    {
                        var guid = Guid.Empty;
                        var result = _endpointVolume.SetMasterVolumeLevelScalar(volume, ref guid);
                        if (result != 0)
                        {
                            // Failed to set, will retry
                            continue;
                        }
                    }
                    catch
                    {
                        // Retry on exception
                        continue;
                    }
                }

                await Task.Delay(delayMs);

                var currentVolume = await GetMasterVolumeAsync();

                if (Math.Abs(currentVolume - volume) < 0.01f)
                {
                    return true;
                }

                await Task.Delay(delayMs);
            }

            return false;
        }

        public async Task<bool> GetMasterMuteAsync()
        {
            return await Task.Run(() =>
            {
                lock (_syncLock)
                {
                    if (_isDisposed || _endpointVolume == null)
                    {
                        return false;
                    }

                    try
                    {
                        var result = _endpointVolume.GetMute(out var mute);
                        return result == 0 ? mute : false;
                    }
                    catch
                    {
                        return false;
                    }
                }
            });
        }

        public async Task<bool> SetMasterMuteAsync(bool mute)
        {
            IAudioEndpointVolume endpointVolumeCopy;
            lock (_syncLock)
            {
                if (_isDisposed || _endpointVolume == null)
                {
                    return false;
                }
                endpointVolumeCopy = _endpointVolume;
            }

            try
            {
                var guid = Guid.Empty;
                return await Task.Run(() => endpointVolumeCopy.SetMute(mute, ref guid) == 0);
            }
            catch
            {
                return false;
            }
        }

        public async Task<List<AudioSession>> GetAudioSessionsAsync()
        {
            return await Task.Run(() =>
            {
                lock (_syncLock)
                {
                    var sessions = new List<AudioSession>();

                    if (_isDisposed || _sessionManager == null)
                    {
                        return sessions;
                    }

                    try
                    {
                        var result = _sessionManager.GetSessionEnumerator(out var sessionEnum);
                        if (result != 0 || sessionEnum == null)
                        {
                            return sessions;
                        }

                        result = sessionEnum.GetCount(out var count);
                        if (result != 0)
                        {
                            return sessions;
                        }

                        for (int i = 0; i < count; i++)
                        {
                            try
                            {
                                result = sessionEnum.GetSession(i, out var sessionControl);
                                if (result == 0 && sessionControl != null)
                                {
                                    sessions.Add(new AudioSession(sessionControl, _sessionManager));
                                }
                            }
                            catch
                            {
                                // Skip on error
                            }
                        }
                    }
                    catch
                    {
                        // Return empty
                    }

                    return sessions;
                }
            });
        }

        public async Task<bool> StepUpAsync(int delayMs = 20)
        {
            bool success = false;
            lock (_syncLock)
            {
                if (_isDisposed || _endpointVolume == null)
                {
                    return false;
                }

                try
                {
                    var guid = Guid.Empty;
                    var result = _endpointVolume.StepUp(ref guid);
                    success = result == 0;
                }
                catch
                {
                    success = false;
                }
            }

            await Task.Delay(delayMs);
            return success;
        }

        public async Task<bool> StepDownAsync(int delayMs = 20)
        {
            bool success = false;
            lock (_syncLock)
            {
                if (_isDisposed || _endpointVolume == null)
                {
                    return false;
                }

                try
                {
                    var guid = Guid.Empty;
                    var result = _endpointVolume.StepDown(ref guid);
                    success = result == 0;
                }
                catch
                {
                    success = false;
                }
            }

            await Task.Delay(delayMs);
            return success;
        }

        public void Dispose()
        {
            lock (_syncLock)
            {
                if (_isDisposed)
                {
                    return;
                }

                ComHelper.ReleaseComObject(_endpointVolume);
                ComHelper.ReleaseComObject(_sessionManager);
                ComHelper.ReleaseComObject(_device);
                _isDisposed = true;
            }
        }
    }
}
